﻿using OpenCvSharp;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ConsoleApp1
{
    class Program
    {
        static void Main(string[] args)
        {
            Action<string, Func<Mat, Mat, Window, int>, int, Mat, Mat, Window> action = (title, work, times, mat, kernel, window) =>
            {
                Stopwatch st = new Stopwatch();
                st.Start();

                Random rand = new Random(Environment.TickCount);

                for (int i = 0; i < times; i++)
                {
                    Point pt1 = new Point(rand.Next(0, mat.Height), rand.Next(0, mat.Width));
                    Point pt2 = new Point(rand.Next(0, mat.Height), rand.Next(0, mat.Width));
                    mat.Line(pt1, pt2, Scalar.Blue);
                    work(mat, kernel, window);
                }

                st.Stop();

                Console.WriteLine(title + " : " + st.ElapsedMilliseconds);
            };

            List<Window> showList = new List<Window>();
            bool doShow = false;

            for (int i = 0; i < 7; i++)
            {
                showList.Add((doShow == true) ? new Window(i.ToString()) : null);
            }

            using (Mat mat = Cv2.ImRead("test_img.png", ImreadModes.GrayScale))
            using (Mat kernel = Mat.Ones(3, 3, (MatType)MatType.CV_8U))
            {
                // kernel.Set(0, 0, 0);

                int count = 10;
                action($"(1)", Convert, 1, mat, kernel, showList[0]);
                action($"opencv({count})", Convert, count, mat, kernel, null);
                Console.WriteLine();

                // Don't use C#: too much slow
                action($"C# for(1)", Convert2, 1, mat, kernel, showList[1]);
                //action($"(1)", Convert2, 1, mat, kernel, null);
                // action($"c# for({count})", Convert2, count, mat, kernel, null);
                Console.WriteLine();

                action($"(1)", Convert3, 1, mat, kernel, showList[2]);
                action($"C++ for({count})", Convert3, count, mat, kernel, null);
                Console.WriteLine();

                action($"(1)", Convert4, 1, mat, kernel, showList[3]);
                action($"C++ parallel for({count})", Convert4, count, mat, kernel, null);
                Console.WriteLine();

                action($"(1)", Convert5, 1, mat, kernel, showList[4]);
                action($"C++ parallel for2({count})", Convert5, count, mat, kernel, null);
                Console.WriteLine();

                action($"(1)", Convert6, 1, mat, kernel, showList[5]);
                action($"C++ custom ({count})", Convert6, count, mat, kernel, null);
                Console.WriteLine();

                action($"(1)", Convert7, 1, mat, kernel, showList[6]);
                action($"C++ custom with simd({count})", Convert7, count, mat, kernel, null);
                Console.WriteLine();
            }

            Console.WriteLine("Press any key to exit...");
            Application.Run();
        }

        static int Convert(Mat mat, Mat kernel, Window window)
        {
            using (Mat erodeMat = mat.Erode(kernel))
            {
                if (window != null)
                {
                    window.ShowImage(erodeMat);
                }

                return erodeMat.Width;
            }
        }

        static unsafe int Convert2(Mat srcMat, Mat kernel, Window window)
        {
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                Parallel.For(0, srcMat.Rows - 2, (y) =>
                {
                    int iMin, iVal;

                    for (int j = 0; j < srcMat.Cols - 2; j++)
                    {
                        iMin = 0xFFF;

                        for (int ii = 0; ii < kernel.Rows; ii++)
                        {
                            for (int jj = 0; jj < kernel.Cols; jj++)
                            {
                                if (kernel.At<byte>(ii, jj) != 0)
                                {
                                    iVal = srcMat.At<byte>(y + ii, j + jj);
                                    if (iMin > iVal)
                                    {
                                        iMin = iVal;
                                    }
                                }
                            }
                        }

                        dstMat.Set<byte>(y + 1, j + 1, (byte)iMin);
                    }
                });

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void erode_cpp_single(byte* srcPtr, int dstRows, int dstCols, byte* kernelPtr, int kernelRows, int kernelCols, byte* dstPtr);

        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void erode_custom(byte* srcPtr, int dstRows, int dstCols, byte* dstPtr);


        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void erode_cpp_parallel(byte* srcPtr, int dstRows, int dstCols, byte* kernelPtr, int kernelRows, int kernelCols, byte* dstPtr);

        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void erode_cpp_parallel2(byte* srcPtr, int dstRows, int dstCols, byte* kernelPtr, int kernelRows, int kernelCols, byte* dstPtr);


        [DllImport("MemCopyLib.dll")]
        public unsafe extern static void erode_custom_parallel(byte* srcPtr, int dstRows, int dstCols, byte* dstPtr);

        static unsafe int Convert3(Mat srcMat, Mat kernel, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                byte* kernelPtr = (byte*)kernel.Data;
                byte* dstPtr = (byte*)dstMat.Data;
                erode_cpp_single(srcPtr, srcMat.Rows, srcMat.Cols, kernelPtr, kernel.Rows, kernel.Cols, dstPtr);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert4(Mat srcMat, Mat kernel, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                byte* kernelPtr = (byte*)kernel.Data;
                byte* dstPtr = (byte*)dstMat.Data;
                erode_cpp_parallel(srcPtr, srcMat.Rows, srcMat.Cols, kernelPtr, kernel.Rows, kernel.Cols, dstPtr);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert5(Mat srcMat, Mat kernel, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                byte* kernelPtr = (byte*)kernel.Data;
                byte* dstPtr = (byte*)dstMat.Data;
                erode_cpp_parallel2(srcPtr, srcMat.Rows, srcMat.Cols, kernelPtr, kernel.Rows, kernel.Cols, dstPtr);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert6(Mat srcMat, Mat kernel, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                byte* kernelPtr = (byte*)kernel.Data;
                byte* dstPtr = (byte*)dstMat.Data;
                erode_custom(srcPtr, srcMat.Rows, srcMat.Cols, dstPtr);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }

        static unsafe int Convert7(Mat srcMat, Mat kernel, Window window)
        {
            byte* srcPtr = (byte*)srcMat.Data;
            using (Mat dstMat = new Mat(srcMat.Rows, srcMat.Cols, (MatType)MatType.CV_8U))
            {
                byte* kernelPtr = (byte*)kernel.Data;
                byte* dstPtr = (byte*)dstMat.Data;
                erode_custom_parallel(srcPtr, srcMat.Rows, srcMat.Cols, dstPtr);

                if (window != null)
                {
                    window.ShowImage(dstMat);
                }
            }

            return 0;
        }
    }
}
